// General-purpose Photovoltaic Device Model gpvdm.com - a drift diffusion
// base/Shockley-Read-Hall model for 1st, 2nd and 3rd generation solarcells.
// The model can simulate OLEDs, Perovskite cells, and OFETs.
// 
// Copyright (C) 2008-2020 Roderick C. I. MacKenzie
// All rights reserved.
// 
// https://www.gpvdm.com
// r.c.i.mackenzie at googlemail.com
// 
// 
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//     * Redistributions of source code must retain the above copyright
//       notice, this list of conditions and the following disclaimer.
//     * Redistributions in binary form must reproduce the above copyright
//       notice, this list of conditions and the following disclaimer in the
//       documentation and/or other materials provided with the distribution.
//     * Neither the name of the GPVDM nor the
//       names of its contributors may be used to endorse or promote products
//       derived from this software without specific prior written permission.
// 
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
// ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
// WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
// DISCLAIMED. IN NO EVENT SHALL Roderick C. I. MacKenzie BE LIABLE FOR ANY
// DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
// (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
// LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
// ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
// SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#ifndef windows
	#include <gsl/gsl_multimin.h>
#endif

#include <sys/types.h>
#include <sys/stat.h>
#include <util.h>
#include <i.h>
#include "../../server.h"
#include "fit.h"
#include "../../dump_ctrl.h"
#include "../../timer.h"

static int unused __attribute__((unused));

static struct fitvars fitconfig;
int uptodate=FALSE;


static double last_constraints_error=0.0;
static double last_total_error=0.0;
static double last_fit_names_error=0.0;

double get_all_error(struct fitvars *myfit)
{
double local_error=0.0;
#ifndef windows
int i;

last_fit_names_error=0.0;

if (myfit->fit_names_len!=0)
{
	for (i=0;i<myfit->fit_names_len;i++)
	{
		myfit->fit_names_error[i]=fit_load_plugin(myfit,myfit->fit_names[i]);
		printf("Last %s error = %le\n",myfit->fit_names[i],myfit->fit_names_error[i]);
		last_fit_names_error+=myfit->fit_names_error[i];
	}

	printf("Total error = %le\n",last_fit_names_error);
}


last_constraints_error=get_constraints_error(myfit);


printf("last_constraints_error = %e\n",last_constraints_error);

last_total_error=last_fit_names_error+last_constraints_error;

printf("total error = %e\n",last_total_error);
#ifndef mindump
FILE *fiterror=fopen("./fiterror.dat","w");
fprintf(fiterror,"%le\n",last_total_error);
fclose(fiterror);
#endif

#endif
return last_total_error;
}

void duplicate()
{
#ifndef windows
int from;
double val;
char orig_file[100];
char dest_file[100];
int orig_pos;
int dest_pos;
char temp[100];
char token1[100];
char token2[100];
if (get_dump_status(dump_iodump)==TRUE) printf("Duplicate\n");
FILE *file;
if ((file=fopen("./duplicate.inp","r"))==NULL)
{
	printf("No duplicate.inp file\n");
	exit(0);
}

do
{
	unused=fscanf(file,"%s",orig_file);
	if (strcmp(orig_file,"#end")==0) break;

	unused=fscanf(file,"%s",token1);
	unused=fscanf(file,"%s",dest_file);
	unused=fscanf(file,"%s",&token2);

	if (orig_file[0]!='#')
	{
		orig_pos=config_search_token(orig_file,token1);
		val=read_value(orig_file,0,orig_pos);
		dest_pos=config_search_token(dest_file,token2);
		edit_file(dest_file,"",dest_pos,val);
	}
}while(!feof(file));
fclose(file);
if (get_dump_status(dump_iodump)==TRUE) printf("Duplicate finished\n");
#endif
}

#ifndef windows
double my_f (const gsl_vector *v, void *params)
{
double e0=0;

if (get_dump_status(dump_iodump)==TRUE) printf("my_f\n");
double val;

int fitcount=0;

for (fitcount=0;fitcount<fitconfig.fitvars;fitcount++)
{
	if (get_dump_status(dump_iodump)==TRUE) printf("fitcount loop %d/%d \n",fitcount,fitconfig.fitvars);
	val = gsl_vector_get(v, fitcount);
	edit_file(fitconfig.fit_file[fitcount],"",fitconfig.fit_pos[fitcount],val);
}

if (get_dump_status(dump_iodump)==TRUE) printf("Exited fitcount loop\n");

duplicate();
e0=fit_run_sims(&fitconfig,&globalserver);


if (get_dump_status(dump_iodump)==TRUE) printf("exit my_f\n");

return e0;
}
#endif

#ifndef windows
void fit_dump_log(int iter,double error,double size,gsl_vector *old_vector,gsl_vector *old_shift,gsl_vector *ss,gsl_vector *x)
{
FILE *out;
if (get_dump_status(dump_iodump)==TRUE) printf("fit_dump_log\n");

gsl_vector_sub (old_vector, x);
gsl_vector_sub (old_shift, ss);

out=fopen("./fitlog.dat","a");
printf ("%d %le %le %le\n", iter, error, last_total_error,size);
fprintf (out,"%d %le %le\n", iter, error, size);
fclose(out);

out=fopen("./fitlog_speed.dat","a");
fprintf (out,"%d %le\n", iter, server_get_jobs_per_s());
fclose(out);

out=fopen("./fitlog_time_error.dat","a");
fprintf (out,"%lf %le\n", timer_get_time(), error);
fclose(out);

out=fopen("./fitlog_time_speed.dat","a");
fprintf (out,"%le %le\n", timer_get_time(), server_get_jobs_per_s());
fclose(out);

out=fopen("./fitlog_time_odes.dat","a");
fprintf (out,"%le %le\n", timer_get_time(), server_get_odes_per_s());
fclose(out);

if (get_dump_status(dump_iodump)==TRUE) printf("exit fit_dump_log\n");

}
#endif

void do_copy()
{
#ifndef windows
FILE *file;
char path[1000];
char command[1000];

	if ((file=fopen("./rsyncdir.inp","r"))==NULL)
	{
		printf("No rsyncdir.inp file so not copying\n");
		exit(0);
	}else
	{

		unused=fscanf(file,"%s",path);
		fclose(file);
		sprintf(command,"cp ./*.dat ./*.inp %s\n",path);

		printf("Running command %s\n",command);
		unused=system(command);

	}

#endif
}

int fit_now(int *oppcount)
{
int cont=TRUE;
#ifndef windows
if (get_dump_status(dump_iodump)==TRUE) printf("fit_now\n");
double val=0.0;
double par[5] = {1.0, 2.0, 10.0, 20.0, 30.0};
size_t iter = 0;
int status;
int item=0;
double size;
int fitcount=0;
int loop=0;
double shift;

fit_read_config(&fitconfig);

shift=fitconfig.simplexmul;
item=fitconfig.fitvars;


gsl_vector *x;
gsl_vector *ss;

gsl_vector *old_vector;
gsl_vector *old_shift;

x = gsl_vector_alloc (fitconfig.fitvars);
ss = gsl_vector_alloc (fitconfig.fitvars);
old_vector = gsl_vector_alloc (fitconfig.fitvars);
old_shift = gsl_vector_alloc (fitconfig.fitvars);

for (fitcount=0;fitcount<fitconfig.fitvars;fitcount++)
{
	val=read_value(fitconfig.fit_file[fitcount],0,fitconfig.fit_pos[fitcount]);
	gsl_vector_set (x, fitcount, val);
	gsl_vector_set (ss, fitcount, fabs(val*shift));
}


	const gsl_multimin_fminimizer_type *T =
	gsl_multimin_fminimizer_nmsimplex;//gsl_multimin_fminimizer_nmsimplex;
	gsl_multimin_function minex_func;
	gsl_multimin_fminimizer *s = NULL;


	/* Initialize method and iterate */
	minex_func.n = item;
	minex_func.f = my_f;
	minex_func.params = par;

	s = gsl_multimin_fminimizer_alloc (T, item);
	gsl_multimin_fminimizer_set (s, &minex_func, x, ss);
	double last_error=0.0;
	int length=0;
	int stop=FALSE;
	int dump_count=0;
	do
	{
		(*oppcount)++;
		if (get_dump_status(dump_iodump)==TRUE)
		{
			printf("call gsl_multimin_fminimizer_iterate\n");
		}
		status = gsl_multimin_fminimizer_iterate(s);
		if (get_dump_status(dump_iodump)==TRUE)
		{
			printf("return from gsl_multimin_fminimizer_iterate\n");
		}

		if (status)
		{
			printf("Exited due to status error\n");
			break;
		}
		size = gsl_multimin_fminimizer_size (s);
		status = gsl_multimin_test_size (size, 1e-8);

		if (status == GSL_SUCCESS)
		{
			printf ("converged to minimum at\n");
		}

		fit_dump_log((*oppcount),s->fval, size,old_vector,old_shift,ss,s->x);

		if (loop>fitconfig.simplexreset)
		{
			if (length>fitconfig.fitvars)
			{
				if (get_dump_status(dump_iodump)==TRUE) printf("one\n");
				stop=TRUE;
			}else
			{
				if (get_dump_status(dump_iodump)==TRUE) printf("two\n");
				loop-=fitconfig.fitvars*2;
			}
		}

		if (last_error==s->fval)
		{
			length++;
		}else
		{
			length=0;
		}

		if (status != GSL_CONTINUE) stop=TRUE;

		if ((dump_count>200)||(stop==TRUE))
		{
			if (get_dump_status(dump_iodump)==TRUE) printf("Calling dump\n");
			do_copy();
			dump_count=0;
		}

		if (get_dump_status(dump_iodump)==TRUE)
		{
			printf("Loop=%d dump_count=%d ittr=%d last_error=%e\n",loop,dump_count,iter,s->fval);
		}

		loop++;
		dump_count++;
		iter++;
		last_error=s->fval;

	}while (stop==FALSE);

	cont=TRUE;

if (get_dump_status(dump_iodump)==TRUE) printf("freeing fit memory\n");
gsl_multimin_fminimizer_free (s);
gsl_vector_free (old_vector);
gsl_vector_free (old_shift);
gsl_vector_free(x);
gsl_vector_free(ss);
if (get_dump_status(dump_iodump)==TRUE) printf("fit_now_exit\n");
#endif
return cont;
}
