// General-purpose Photovoltaic Device Model gpvdm.com - a drift diffusion
// base/Shockley-Read-Hall model for 1st, 2nd and 3rd generation solarcells.
// The model can simulate OLEDs, Perovskite cells, and OFETs.
// 
// Copyright (C) 2008-2020 Roderick C. I. MacKenzie
// All rights reserved.
// 
// https://www.gpvdm.com
// r.c.i.mackenzie at googlemail.com
// 
// 
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//     * Redistributions of source code must retain the above copyright
//       notice, this list of conditions and the following disclaimer.
//     * Redistributions in binary form must reproduce the above copyright
//       notice, this list of conditions and the following disclaimer in the
//       documentation and/or other materials provided with the distribution.
//     * Neither the name of the GPVDM nor the
//       names of its contributors may be used to endorse or promote products
//       derived from this software without specific prior written permission.
// 
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
// ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
// WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
// DISCLAIMED. IN NO EVENT SHALL Roderick C. I. MacKenzie BE LIABLE FOR ANY
// DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
// (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
// LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
// ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
// SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "../../sim.h"
#include "../../dump.h"
#include "../../ntricks.h"
#include "tpc.h"
#include "dump.h"
#include "inp.h"
#include "buffer.h"
#include "light_interface.h"
#include "../../dynamic_store.h"

struct tpc tpc_config;

static int unused __attribute__((unused));

void sim_tpc(struct device *in)
{
int i;
struct buffer buf;
struct istruct tpcfitout;

buffer_init(&buf);

tpc_load_config(&tpc_config,in);

in->dt=tpc_config.tpcdtstart;

in->go_time=FALSE;


inter_init(&tpcfitout);

struct dynamic_store store;
dump_dynamic_init(&store,in);

printf("Running tpc. %e\n",tpc_config.tpc_externalv);
double V=0.0;


int step=0;

light_solve_and_update(in,&(in->mylight),in->Psun,0.0,0.0);
if (fabs(tpc_config.tpc_externalv-in->Vapplied)>0.2)
{
	ramp_externalv(in,0.0,tpc_config.tpc_externalv);
}

sim_externalv(in,tpc_config.tpc_externalv);

tpc_config.tpc_power_suply=tpc_config.tpc_externalv+get_equiv_I(in)*tpc_config.Rtpc_short;

time_init(in);
device_timestep(in);

in->go_time=TRUE;
V=in->Vapplied;

in->time=tpc_config.tpc_time_start;
double mue;
double muh;
double ratio;
double Psun_orig=in->Psun;

if (in->dumpitdos==TRUE)
{
	mue=get_avg_mue(in);
	muh=get_avg_muh(in);
}
carrier_count_reset(in);

do
{


	if (in->time<tpc_config.tpc_time_start)
	{
		light_solve_and_update(in,&(in->mylight),in->Psun,0.0,0.0);
		reset_np_save(in);
	}else
	if ((in->time>=tpc_config.tpc_pulse_start)&&(in->time<(tpc_config.tpc_pulse_start+tpc_config.tpc_laser_pulse_width)))
	{

		ratio=light_convert_density(in,tpc_config.tpc_pulse_start,tpc_config.tpc_laser_pulse_width);
		if (ratio>0.0)
		{
			light_solve_and_update(in,&(in->mylight),in->Psun,ratio*tpc_config.tpc_laser_eff,tpc_config.tpc_laser_pulse_width);
		}else
		{
			light_solve_and_update(in,&(in->mylight),in->Psun,0.0,0.0);
		}

	}else
	{
		light_solve_and_update(in,&(in->mylight),in->Psun,0.0,0.0);
	}

	in->Vapplied=V;


	if (tpc_config.tpc_sim_all==TRUE)
	{
		double i0=newton_sim_tpc_all(in);
		inter_append(&tpcfitout,in->time,i0);
	}else
	{
		newton_sim_tpc(in);
		inter_append(&tpcfitout,in->time,get_equiv_I(in));
	}

	if (get_dump_status(dump_print_converge)==TRUE)
	{
		printf("tpc time=%e step=%d %.1e %d ",in->time,step,in->last_error,in->last_ittr);
		printf("V=%lf current = %e mA (%e A/m^2)\n",V,get_I(in)/1e-3,get_J(in));
	}

	dump_write_to_disk(in);


	plot_now(in,in->plot_file);
	dump_dynamic_add_data(&store,in,in->time);

	if (in->time>(tpc_config.tpc_pulse_start+tpc_config.tpc_laser_pulse_width)) in->dt*=tpc_config.tpcdtmull;

	device_timestep(in);
	stop_start(in);
	step++;

	if (in->time>tpc_config.tpcstop) break;

	if (in->stop==TRUE)
	{
		break;

	}


}while(1);

in->go_time=FALSE;

char name[200];

inter_sub_double(&tpcfitout,tpcfitout.data[tpcfitout.len-1]);
inter_rescale(&tpcfitout,1.0, -1.0);
inter_mod(&tpcfitout);

buffer_malloc(&buf);
sprintf(name,"%s%s","tpc_i_no_dc",".dat");
buf.y_mul=1.0;
buf.x_mul=1.0;
strcpy(buf.title,"Transient Photo Current");
strcpy(buf.type,"xy");
strcpy(buf.x_label,"Time");
strcpy(buf.y_label,"Current");
strcpy(buf.x_units,"s");
strcpy(buf.y_units,"A");
buf.logscale_x=1;
buf.logscale_y=1;
buffer_add_info(&buf);
buffer_add_xy_data(&buf,tpcfitout.x, tpcfitout.data, tpcfitout.len);
buffer_dump(in->outputpath,name,&buf);
buffer_free(&buf);



#ifndef mindump
double half_time=inter_get_quartile(&tpcfitout,0.5);



if (in->dumpitdos==TRUE)
{


	FILE *half=fopen("./tpctransittime.dat","w");
	fprintf(half,"%le\n",half_time);
	fclose(half);

	half=fopen("./tpcmobility.dat","w");
	fprintf(half,"%le %le %le %le %le\n",mue,muh,in->mun[0],in->mup[0],(in->ylen*in->ylen)/(half_time*2.0*fabs(in->phi[0]-in->phi[in->ymeshpoints-1])));
	fclose(half);

	inter_chop(&tpcfitout,1e-9, 1.0);

	for (i=0;i<tpcfitout.len;i++)
	{
	tpcfitout.data[i]*=tpcfitout.x[i];
	tpcfitout.x[i]=((in->Te[0]*kb/Q)*log(1e7*tpcfitout.x[i]));
	}

	double val=inter_get_hard(&tpcfitout,0.0);

	inter_rescale(&tpcfitout,1.0, 1.0/val);

	sprintf(name,"%s%s",in->outputpath,"tpcitdos.dat");
	inter_save(&tpcfitout,name);

}

dump_dynamic_save(in->outputpath,&store);
dump_dynamic_free(&store);
#endif


inter_free(&tpcfitout);
}

void tpc_load_config(struct tpc *in,struct device *dev)
{
struct inp_file inp;
inp_load(&inp,dev->outputpath,"tpc.inp");
inp_check(&inp,1.11);
inp_search_double(&inp,&(in->tpcstop),"#tpc_stop");
inp_search_double(&inp,&(in->tpc_laser_pulse_width),"#tpc_laser_pulse_width");
inp_search_double(&inp,&(in->tpcdtstart),"#tpc_dtstart");
inp_search_double(&inp,&(in->tpcdtmull),"#tpc_dtmull");
inp_search_double(&inp,&(in->tpc_laser_eff),"#tpc_laser_eff");
inp_search_double(&inp,&(in->Rtpc_short),"#tpc_short");
inp_search_int(&inp,&(in->tsteps),"#tpc_tsteps");
inp_search_double(&inp,&(in->tpc_pulse_start),"#tpc_pulse_start");
inp_search_double(&inp,&(in->tpc_time_start),"#tpc_time_start");
inp_search_int(&inp,&(in->reprate),"#tpc_reprate");
inp_search_double(&inp,&(in->timeshift),"#tpc_timeshift");
inp_search_double(&inp,&(tpc_config.tpc_externalv),"#tpc_externalv");
inp_search_int(&inp,&(in->tpc_sim_all),"#tpc_simall");
inp_search_double(&inp,&(in->tpc_L),"#tpc_L");
inp_free(&inp);
}
